=== modified file 'bzrlib/dirstate.py'
--- bzrlib/dirstate.py	2007-02-12 02:18:44 +0000
+++ bzrlib/dirstate.py	2007-02-12 05:09:17 +0000
@@ -245,13 +245,13 @@
         to_yesno = DirState._to_yesno
 
         st = os.lstat(tree.basedir)
-        root_info = [
+        root_info = (
             '', '' # No path
             , 'directory', tree.inventory.root.file_id.encode('utf8')
             , 0 # no point having a size for dirs.
             , pack_stat(st)
             , '' # No sha
-            ]
+            )
         root_parents = []
         for parent_tree in parent_trees:
             root_parents.append((
@@ -294,7 +294,7 @@
                 parent_info = []
                 for count in xrange(num_parents):
                     parent_info.append(
-                        self._parent_info(parent_trees[count], fileid))
+                        result._parent_info(parent_trees[count], fileid))
                 row_data = (dirname.encode('utf8'), basename.encode('utf8'),
                     kind, fileid.encode('utf8'), st.st_size, pack_stat(st),
                     s)
@@ -478,17 +478,20 @@
 
             assert len(entries) == self._num_entries, '%s != %s entries' % (len(entries),
                 self._num_entries)
-            entry_iter = iter(entries)
-            self._root_row = entry_iter.next()
-            # convert the minikind to kind
-            self._root_row[0][2] = self._minikind_to_kind[self._root_row[0][2]]
-            # convert the size to an int
-            self._root_row[0][4] = int(self._root_row[0][4])
-            # TODO parent converion
-            # TODO dirblock population
-            for entry in entry_iter:
-                # do something here
-                pass
+
+            def _line_to_row(line):
+                """Convert a freshly read line's size and minikind for use."""
+                # convert the minikind to kind
+                line[0][2] = self._minikind_to_kind[line[0][2]]
+                # convert the size to an int
+                line[0][4] = int(line[0][4])
+                for parent in line[1]:
+                    parent[1] = self._minikind_to_kind[parent[1]]
+                    parent[4] = int(parent[4])
+                    parent[5] = parent[5] == 'y'
+                return tuple(line[0]), map(tuple, line[1])
+            new_rows = map(_line_to_row, entries)
+            self._rows_to_current_state(new_rows)
         self._dirblock_state = DirState.IN_MEMORY_UNMODIFIED
 
     def _read_header(self):
@@ -557,6 +560,28 @@
         entire_row[4] = str(entire_row[4])
         # minikind of parents
         return '\0'.join(entire_row)
+
+    def _rows_to_current_state(self, new_rows):
+        """Load new_rows into self._root_row and self.dirblocks.
+
+        Process new_rows into the current state object, making them the active
+        state.
+
+        :param new_rows: A sorted list of rows. This function does not sort
+            to prevent unneeded overhead when callers have a sorted list
+            already.
+        :return: Nothing.
+        """
+        assert new_rows[0][0][0:2] == ('', ''), \
+            "Incorrect root row %r" % new_rows[0][0]
+        self._root_row = new_rows[0]
+        self._dirblocks = [('', [])]
+        for row in new_rows[1:]:
+            if row[0] != self._dirblocks[-1][0]:
+                # new block
+                self._dirblocks.append((row[0][0], []))
+            # append the row to the current block
+            self._dirblocks[-1][1].append(row)
     
     def save(self):
         """Save any pending changes created during this session."""
@@ -634,17 +659,39 @@
         # skip ghost trees, as they dont get represented.
         parent_trees = [tree for rev_id, tree in trees if rev_id not in ghosts]
         parent_tree_count = len(parent_trees)
+        # loop over existing entries in the dirstate.
+        checked_ids = set()
         for entry, old_parents in self._iter_rows():
             file_id = entry[3]
-            new_parents = [None] * parent_tree_count
-            for position, parent_tree in enumerate(parent_trees):
-                # revision_utf8, KIND, dirname, basename, size, executable, sha
-                new_parents[position] = self._parent_info(parent_tree, file_id)
-            assert None not in new_parents
-            new_rows.append((entry, new_parents))
-        self._root_row = new_rows[0]
-        self._dirblocks = []
-         
+            checked_ids.add(file_id)
+            new_parents = [None] * parent_tree_count
+            for position, parent_tree in enumerate(parent_trees):
+                # revision_utf8, KIND, dirname, basename, size, executable, sha
+                new_parents[position] = self._parent_info(parent_tree, file_id)
+            assert None not in new_parents
+            new_rows.append((entry, new_parents))
+        # get additional ids that are present in parents and not in this tree.
+        deleted_ids = set()
+        for tree in parent_trees:
+            deleted_ids.update(set(tree.inventory._byid).difference(checked_ids))
+        # add the deleted ids to the dirstate. deleted files are represented as
+        # a file with dirname '', basename ''
+        for file_id in deleted_ids:
+            # add these ids to the deleted block
+            checked_ids.add(file_id)
+            # deleted items have a synthetic entry.
+            entry = ('/', 'RECYCLED.BIN', 'file', file_id.encode('utf8'), 0,
+                DirState.NULLSTAT, '')
+            new_parents = [None] * parent_tree_count
+            for position, parent_tree in enumerate(parent_trees):
+                # revision_utf8, KIND, dirname, basename, size, executable, sha
+                new_parents[position] = self._parent_info(parent_tree, file_id)
+            assert None not in new_parents
+            new_rows.append((entry, new_parents))
+
+        # sort all the rows
+        new_rows = sorted(new_rows)
+        self._rows_to_current_state(new_rows)
         self._parents = [rev_id for rev_id, tree in trees]
         self._ghosts = list(ghosts)
         self._header_state = DirState.IN_MEMORY_MODIFIED

=== modified file 'bzrlib/tests/test_dirstate.py'
--- bzrlib/tests/test_dirstate.py	2007-02-12 02:18:44 +0000
+++ bzrlib/tests/test_dirstate.py	2007-02-12 05:09:17 +0000
@@ -19,6 +19,7 @@
 import os
 
 from bzrlib import dirstate
+from bzrlib.dirstate import DirState
 from bzrlib.memorytree import MemoryTree
 from bzrlib.tests import TestCaseWithTransport
 
@@ -62,7 +63,7 @@
             # there should be one fileid in this tree - the root of the tree.
             root_stat_pack = dirstate.pack_stat(os.stat(tree.basedir))
             self.assertEqual(
-                [(['', '', 'directory', tree.inventory.root.file_id, 0, root_stat_pack, ''], [])],
+                [(('', '', 'directory', tree.inventory.root.file_id, 0, root_stat_pack, ''), [])],
                 list(state._iter_rows()))
         check_state()
         state = dirstate.DirState.on_file('dirstate')
@@ -314,11 +315,51 @@
         # the ghost should be recorded as such by set_parent_trees.
         self.assertEqual(['ghost-rev'], state.get_ghosts())
         self.assertEqual(
-            [(['', '', 'directory', root_id, 0, 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', ''], [
+            [(('', '', 'directory', root_id, 0, 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx', ''), [
              (revid1, 'directory', '', '', 0, False, ''),
              (revid2, 'directory', '', '', 0, False, '')])],
             list(state._iter_rows()))
 
+    def test_set_parent_trees_file_missing_from_tree(self):
+        # Adding a parent tree may reference files not in the current state.
+        # they should get listed just once by id, even if they are in two  
+        # separate trees.
+        # set_parent_trees is a slow but important api to support.
+        state = dirstate.DirState.initialize('dirstate')
+        tree1 = self.make_branch_and_memory_tree('tree1')
+        tree1.lock_write()
+        tree1.add('')
+        tree1.add(['a file'], ['file-id'], ['file'])
+        tree1.put_file_bytes_non_atomic('file-id', 'file-content')
+        revid1 = tree1.commit('foo')
+        tree1.unlock()
+        branch2 = tree1.branch.bzrdir.clone('tree2').open_branch()
+        tree2 = MemoryTree.create_on_branch(branch2)
+        tree2.lock_write()
+        tree2.put_file_bytes_non_atomic('file-id', 'new file-content')
+        revid2 = tree2.commit('foo')
+        root_id = tree2.inventory.root.file_id
+        state.set_path_id('', root_id)
+        tree2.unlock()
+        state.set_parent_trees(
+            ((revid1, tree1.branch.repository.revision_tree(revid1)),
+             (revid2, tree2.branch.repository.revision_tree(revid2)),
+             ), [])
+        # check the layout in memory
+        expected_rows = [
+            (('', '', 'directory', root_id, 0, DirState.NULLSTAT, ''),
+             [(revid1.encode('utf8'), 'directory', '', '', 0, False, ''),
+              (revid2.encode('utf8'), 'directory', '', '', 0, False, '')]),
+            (('/', 'RECYCLED.BIN', 'file', 'file-id', 0, DirState.NULLSTAT, ''),
+             [(revid1.encode('utf8'), 'file', '', 'a file', 12, False, '2439573625385400f2a669657a7db6ae7515d371'),
+              (revid2.encode('utf8'), 'file', '', 'a file', 16, False, '542e57dc1cda4af37cb8e55ec07ce60364bb3c7d')])
+            ]
+        self.assertEqual(expected_rows, list(state._iter_rows()))
+        # check we can reopen and use the dirstate after setting parent trees.
+        state.save()
+        state = dirstate.DirState.on_file('dirstate')
+        self.assertEqual(expected_rows, list(state._iter_rows()))
+
     ### add a path via _set_data - so we dont need delta work, just
     # raw data in, and ensure that it comes out via get_lines happily.
 

